import logging
import asyncio
import uuid
import time
import re

from rasa.core.channels.channel import UserMessage
from typing import List, Callable

logger = logging.getLogger(__name__)


class MiddlewareConnector:

    """
    Abstract class that adds middleware capacity to the connector.

    To instantiate this class you should implement 'get_middlewares', 
    'get_on_new_message' and 'create_user_message'.
    """

    used_middlewares = []
    middleware_is_ready = False

    def get_middlewares(self) -> List:
        """
        This method should return a list with your middleware objects 
        (created inheriting from `BaseMiddleware` class or that are compatible 
        with the same interface) in order of execution. 

        The Rasa Agent handler will be added after the middlewares.
        
        If no middleware is provided (returnin a empty list), the message will 
        be send directly to Rasa.
        """

        raise NotImplementedError

    def get_on_new_message(self) -> Callable:
        """
        This method returns the Rasa Agent `handle_message` endpoint, 
        usually it is received as the parameter `on_new_message` on the `blueprint` 
        function of a handler.

        It will be used to send the message to Rasa after all middlewares finish.
        """

        raise NotImplementedError

    def create_user_message(self, *args, **kwargs) -> UserMessage:
        """
        Returns a UserMessage object containing the text to be processed.
        """
        
        raise NotImplementedError

    def setup_middlewares(self):
        """
        This method goes trough the middlewares supplied on the 
        get_middlewares function and sets the 'next' attribute following 
        the order of the list. Appens the on_new_message (Rasa Agent Handler) 
        to the end of the middleware list. 
        """
        
        last = None
        for middleware in self.get_middlewares():
            
            if last is not None:
                last.set_next(middleware.compute)

            last = middleware
            self.used_middlewares.append(middleware)


        if len(self.used_middlewares) > 0:
            last.set_next(self.get_on_new_message())
        else:
            self.used_middlewares = [self.get_on_new_message(), ]
    
        self.middleware_is_ready = True

    async def proccess_message(self, message: UserMessage):
        """
        Send the message object to the first middleware, starting the 
        processment stage.
        """

        await self.used_middlewares[0].compute(message)

    async def handle_message(self, *args, **kwargs):
        """
        Method that handles a new message beeing sent to the connector.
        """

        # setup middlewares it not ready
        if not self.middleware_is_ready:
            self.setup_middlewares()

        # get a UserMessage object from args passed
        message = self.create_user_message(*args, **kwargs)
        
        # sends UserMessage to middlewares
        await self.proccess_message(message)

    






import asyncio
import logging

from rasa.core.channels.channel import UserMessage

from rasa_middleware_connector import BaseMiddleware

logger = logging.getLogger(__name__)

class TextCleaner(BaseMiddleware):

    """
    Cleans message from selected expressions.
    """

    async def compute(self, message: UserMessage):

        logger.info("Middleware TextCleaner received message from {}".format(message.sender_id))
        
        message.text = self.clean_message(message.text)
        await self.next(message)

    def clean_message(self, text: str):

            text = text.strip()
            text = text.lower()

            replacements = {
                'a': ['à', 'á', 'ã', 'ä'],
                'e': ['ê', 'ẽ', 'è', 'ë', 'eh', 'é'],
                'i' : ['í', 'ì', 'î', 'ĩ'],
                'o': ['ó', 'ò', 'õ', 'ö'],
                'u': ['ú', 'ù', 'ũ', 'ü'],
                'c': ['ç'],
                'voce': ['vc'],
                '': [','],
                'tambem': ['tbm'],
                'hoje': ['hj'],
                'tudo': ['td'],
                ' esta ': [' ta '],
                ' para ': [' pra ']
            }

            for key, lst in replacements.items():
                for c in lst:
                    text = text.replace(c, key)

            return text


import logging
import asyncio

from asyncio import Lock
from rasa.core.channels.channel import UserMessage

from rasa_middleware_connector import BaseMiddleware

logger = logging.getLogger(__name__)

class MessageCollector(BaseMiddleware):

    """
    Collects messages sent during a time periodo and combine them 
    into a single message.
    """
    
    mutex = Lock()
    __on_new_message = None

    def __init__(self, connector):
        self.handlers = {}
        self.connector = connector
        
    async def compute(self, message: UserMessage):

        logger.info("Middleware MessageCollector received message from {}".format(message.sender_id))
                
        await self.register_message(message, self.next)   

    async def register_message(self, user_message: UserMessage, on_new_message):
        
        sid = user_message.sender_id
        
        if sid in self.handlers:

            try:
                await self.handlers[sid].append_message(user_message)
                logger.info("Added menssage to handler:" + sid)
            except HandlerClosedException:
                await self.create_handler(user_message, on_new_message)

        else:
            # skips delay on first message
            await self.create_handler(user_message, on_new_message, 0)

    async def create_handler(self, user_message, on_new_message, delay=None):
        async with self.mutex:

            sid = user_message.sender_id

            logger.info("Creating handler for: " + sid)

            new_handler = MessageHandler(user_message, on_new_message, delay=delay)
            self.handlers[sid] = new_handler
            

class MessageHandler:
    
    def __init__(self, user_message: UserMessage, on_new_message, delay=None):
        from os import getenv

        if delay is None:
            self.DELAY = float(getenv('DELAY_TIME', 3))
        else:
            self.DELAY = float(delay)

        self.messages = [user_message, ]
        self.sid = user_message.sender_id
        self.on_new_message = on_new_message

        self.mutex = Lock()
        self.reached_commit = False
        self.accepting = True
        

        logger.info(
            "Initializing handler for {} with delay of {}s and message {}".format(
                self.sid, 
                self.DELAY,
                user_message.text
            )
        )
        
        self.timer = AsyncTimer(self.DELAY, self.commit)

    async def append_message(self, user_message: UserMessage):
        
        async with self.mutex:
            if self.accepting is False:
                raise HandlerClosedException()

            full_text = user_message.text
            text_parts = await self.preprocess_text(full_text)
            
            self.messages.append(user_message)
            self.__reset_timer()
            
    def __reset_timer(self):

        self.timer.cancel()
        
        # if the commit has been reached, we shoud not reset the
        # timer, or it will process the message twice
        # there is no problem in not reseting the timer, since the message
        # added in this append operation will still be included
        # thats because the commit function awaits the lock that the funcion 
        # append (caller of this function) has on the message list
        # when the appen function releases the lock, all messages that need
        # to be processed will be alreadly included in the list
        if self.reached_commit is False:
            
            self.timer = AsyncTimer(self.DELAY, self.commit)
  
    async def commit(self):

        # signals that the timer has called this function
        self.reached_commit = True
        logger.info("Handler {} reached commit".format(self.sid))

        async with self.mutex:

            logger.info("Handler {} acquired lock for commit".format(self.sid))

            # close this Handler
            self.accepting = False
            
            # append all messages with a space in between
            complete_text = self.messages[0].text
            for msg in self.messages[1:]:
                complete_text += ' ' + msg.text

            final_message = self.messages[0]
            final_message.text = complete_text

            logger.info("Handler {} finished, sending final message '{}'".format(self.sid, complete_text))

            await self.on_new_message(final_message)
            logger.info("Handler {} closed with messages: {}".format(self.sid, self.messages))
                       
            
class AsyncTimer:
    def __init__(self, timeout, callback):
        self._timeout = timeout
        self._callback = callback
        self._task = asyncio.ensure_future(self._job())

    async def _job(self):
        await asyncio.sleep(self._timeout)
        await self._callback()

    def cancel(self):
        self._task.cancel()

class HandlerClosedException(Exception):
    pass


from rasa.core.channels.channel import UserMessage
from typing import Callable

class BaseMiddleware:

    """
    Interface for Middlewares
    """

    next = None

    def set_next(self, next: Callable[[UserMessage], None]):

        """
        Sets the next middleware to call.
        """

        self.next = next

    async def compute(self, message: UserMessage):

        """
        This method is where the UserMessage will be processed.

        When done, 'await self.next(message)' should be called.
        """

        # await self.next(message)
        raise NotImplementedError()



import logging
import asyncio

from rasa.core.channels.channel import UserMessage
from rasa.core.channels.socketio import SocketIOInput, SocketIOOutput

from rasa_middleware_connector import MiddlewareConnector
from .custom_middlewares.message_collector import MessageCollector
from .custom_middlewares.text_cleaner import TextCleaner

logger = logging.getLogger(__name__)

class SocketInput(SocketIOInput, MiddlewareConnector):

    """A socket.io input channel with middleware support."""

    on_new_message = None
    sio = None

    def get_middlewares(self):

        # returns the middleware list that this connector will use
        # they will be executed respecting the list order and the
        # rasa 'on_new_message' will be added as the last element

        # if this list is empty, only 'on_new_message' will be called
        
        return [
            TextCleaner(),
            MessageCollector(self)        
        ]

    def get_on_new_message(self):

        # returns the default rasa handler for messages
        # this function sends the message objecto to the
        # rasa agent that will handle it

        return self.on_new_message

    def create_user_message(self, sid, data) -> UserMessage:

        # this function receives required data to create a new
        # user message and returns it (based on the official rasa
        # implementation of the socket connector) 
      
        if self.session_persistence:
            if not data.get("session_id"):
                logger.warning(
                    "A message without a valid sender_id "
                    "was received. This message will be "
                    "ignored. Make sure to set a proper "
                    "session id using the "
                    "`session_request` socketIO event."
                )
                return
            sender_id = data["session_id"]
        else:
            sender_id = sid

        text = data["message"]

        output_channel = SocketIOOutput(self.sio, sender_id, self.bot_message_evt)
        message = UserMessage(
            text, output_channel, sid, input_channel=self.name()
        )
        
        return message

    def blueprint(self, on_new_message):

        self.on_new_message = on_new_message

        socketio_webhook = super().blueprint(on_new_message)
        self.sio = socketio_webhook.sio

        # swaps default event handler with the desired handler
        self.sio.handlers = socketio_webhook.sio.handlers
        self.sio.handlers['/'][self.user_message_evt] = self.handle_message

        return socketio_webhook

